<?php

namespace Drupal\interactive_queue_processing;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\Queue\QueueWorkerManagerInterface;
use Drupal\Core\Queue\RequeueException;
use Drupal\Core\Queue\SuspendQueueException;

/**
 * Various queue management functionality.
 */
class QueueManager {

  /**
   * The queue factory service.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueFactory;

  /**
   * The queue worker plugin manager service.
   *
   * @var \Drupal\Core\Queue\QueueWorkerManagerInterface
   */
  protected $queueWorkerManager;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * QueueManager constructor.
   *
   * @param \Drupal\Core\Queue\QueueFactory $queueFactory
   *   The queue factory service.
   * @param \Drupal\Core\Queue\QueueWorkerManagerInterface $queueWorkerManager
   *   The queue worker plugin manager service.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   */
  public function __construct(QueueFactory $queueFactory, QueueWorkerManagerInterface $queueWorkerManager, TimeInterface $time) {
    $this->queueFactory = $queueFactory;
    $this->queueWorkerManager = $queueWorkerManager;
    $this->time = $time;
  }

  /**
   * Gets the queue definitions.
   *
   * @return array
   *   Queue definitions.
   */
  public function getQueueDefinitions() {
    return array_filter($this->queueWorkerManager->getDefinitions(), function ($info) {
      return isset($info['cron']);
    });
  }

  /**
   * Processes cron queues.
   *
   * @see \Drupal\Core\Cron::processQueues
   */
  public function processQueues() {
    // We'll only process for ~10 seconds.
    $start = $this->time->getCurrentMicroTime();
    $end = $start + 10;

    foreach ($this->getQueueDefinitions() as $queue_name => $info) {
      $this->queueFactory->get($queue_name)->createQueue();

      $queue_worker = $this->queueWorkerManager->createInstance($queue_name);
      $queue = $this->queueFactory->get($queue_name);
      $lease_time = isset($info['cron']['time']) ?: NULL;

      while ($this->time->getCurrentMicroTime() < $end && ($item = $queue->claimItem($lease_time))) {
        try {
          $queue_worker->processItem($item->data);
          $queue->deleteItem($item);
        }
        catch (RequeueException $e) {
          // The worker requested the task be immediately requeued.
          $queue->releaseItem($item);
        }
        catch (SuspendQueueException $e) {
          // If the worker indicates there is a problem with the whole queue,
          // release the item and skip to the next queue.
          $queue->releaseItem($item);

          watchdog_exception('interactive_queue_processing', $e);

          // Skip to the next queue.
          continue 2;
        }
        catch (\Exception $e) {
          // In case of any other kind of exception, log it and leave the item
          // in the queue to be processed again later.
          watchdog_exception('interactive_queue_processing', $e);
        }
      }
    }
  }

  /**
   * Counts items in queues.
   *
   * @return array
   *   Number of items in each queue, keyed by queue name.
   */
  public function queueCounts() {
    $count = [];

    foreach ($this->getQueueDefinitions() as $info) {
      $queue = $this->queueFactory->get($info['id']);
      $count[$info['id']] = $queue->numberOfItems();
    }

    return $count;
  }

}
